#include "dynamics.h"

#include <Eigen/LU>
#include <iostream>
#include <stdlib.h>
#include <fstream>
#include <cmath>

#include "../settings.h"

using namespace std;
using namespace imr::dynamics;

/**
 * Constructor of the class Dynamics.
 * @param initial_X     - initial position of the quadcopter
 * @param initial_V     - initial velocity of the quadcopter
 * @param initial_OMEGA - initial angular velocity of the quadcopter
 * @param initial_R     - initial rotation matrix of the quadcopter
 * @param type          - 1 - leader = control history is stored; 0 - follower = control history is not stored
 * @param filename      - file to save the simulation results
 * */
Dynamics::Dynamics(const Vector & initial_X, const Vector & initial_V, const Vector & initial_OMEGA,
                   	   const Matrix & initial_R, const int _type, const char * filename) :
                   	   	   Quadrotor(initial_X, initial_V, initial_OMEGA, initial_R, _type)
{
    log_file = filename;
    traveled_distance.push_back(0.0);
}

/**
 * Moves the quadrotor to desired state.
 * @param ctr - control (vt,vn,k,t)
 */
double Dynamics::move_ctr(const OptVec & ctr)
{
	double t = ctr.t;
	double _t = 0.0;

	// Normal acceleration
	double an = 0.0;
	// Tangential acceleration
	double at = 0.0;
	// Centripetal acceleration
	double ac = 0.0;
	// Total thrust
	double f = 0.0;

	const double vt = state.V(1)*sin(state.EA(2)) + state.V(0)*cos(state.EA(2));

		#ifdef DEBUG_DYNAMICS
		cout << "\n\n\n\nrobot_move" << endl;
		cout << "EA = (" << state.EA.transpose() << ")'" << endl;
		cout << "V = (" << state.V.transpose() << ")'" << endl;
		cout << "vt = " << vt << endl;
		cout << "om = " << ctr.vt*ctr.c << endl;
		cout << "W = (" << state.W.transpose() << ")'" << endl;
		#endif

	const double delta_vt = ctr.vt - vt;
	const double delta_vn = ctr.vn - state.V(2);

	// accelerate / decelerate
	if (abs(delta_vt) > V_TOL_XY || abs(delta_vn) > V_TOL_Z) {
		if (delta_vt > V_TOL_XY) {
			at = AT_MAX;
			_t = min(t, delta_vt/at);
		} else if (delta_vt < -V_TOL_XY) {
			at = AT_MIN;
			_t = min(t, abs(delta_vt/at));
		}
		if (_t == 0.0) {
			if (delta_vn > V_TOL_Z) {
				an = AN_MAX;
				_t = min(t, delta_vn/an);
			} else if (delta_vn < -V_TOL_Z){
				an = AN_MIN;
				_t = min(t, abs(delta_vn/an));
			}
		} else {
			if (delta_vn > constants::eps) {
				an = min(delta_vn/_t, AN_MAX);
			} else if (delta_vn < -constants::eps){
				an = max(delta_vn/_t, AN_MIN);
			}
		}

		const double psi = atan2(at, an+constants::g); // pitch
			#ifdef DEBUG_DYNAMICS
			cout << "ACCELERATE" << endl;
			cout << "psi = " << psi << endl;
			#endif

		f = (m*(an+constants::g))/cos(psi);

		if (abs(state.EA(1) - psi) > A_TOL || abs(state.EA(0)) > A_TOL) {
			tilt(-state.EA(0), psi - state.EA(1), f, TURN_T);
			_t -= TURN_T;
			t -= TURN_T;
		}

		if (_t > constants::eps) {
			static_control_move(_t, f, Eigen::Vector3d::Zero());
			t -= _t;
		}
	}

	if (t > TURN_T) {
		ac = sqr(ctr.vt)*ctr.c;

		const double phi = atan2(-ac, constants::g); // roll
			#ifdef DEBUG_DYNAMICS
			cout << "TURN" << endl;
			cout << "phi = " << phi << endl;
			#endif

		f = (m*(constants::g))/cos(phi);

		if (abs(state.EA(0) - phi) > A_TOL || abs(state.EA(1)) > A_TOL) {
			tilt(phi - state.EA(0), -state.EA(1), f, TURN_T);
			t -= TURN_T;
		}

		if (t > constants::eps) {
			turn(ctr.vt*ctr.c, f, t);
		}
	}
	return traveled_distance.back();
}

/**
 * Function moves the robot by given roll and pitch Euler angles.
 * @param roll - angle increment (around x-axis)
 * @param pitch - angle increment (around y-axis)
 * @param f - total thrust
 * @param t - time
 * */
void Dynamics::tilt(const double roll, const double pitch, const double f, const double t)
{
	const double tt = t-3*DELTA_T;
	const Vector W(roll/tt, pitch/tt, 0.0);
	const Vector M = (J*W)/DELTA_T;

	/// Stop quadrotor rotation
	static_control_move(DELTA_T, f, calculate_moment_inputs(Eigen::Vector3d::Zero()));

	/// Tilt the quadrotor to given euler angles
	static_control_move(DELTA_T, f, M);
	static_control_move(tt, f, Eigen::Vector3d::Zero());
	static_control_move(DELTA_T, f, -M);
}

/**
 * Function keeps given angular velocity with respect to quadrotor's body-fixed frame.
 * @param om - angular velocity
 * @param t  - control time
 * */
void Dynamics::turn(const double om, const double f, const double t)
{
	Vector W = state.R.inverse()*Eigen::Vector3d::UnitZ();
	W *= om;
	static_control_move(DELTA_T, f, calculate_moment_inputs(W));
	static_control_move(t-DELTA_T,f,W.cross(J*W));
}

/**
 * Function calculating the moment input vector.
 * @param W - desired angular velocity vector
 * @return M - moment inputs
 */
const Vector Dynamics::calculate_moment_inputs(const Vector W) {
	Vector M;
	M = J*W;
	M.noalias() -= J*state.W;
	M.noalias() += state.W.cross(J*state.W);
	M /= DELTA_T;
	return M;
}

State Dynamics::get_leaders_state(const double s, int v) const
{
	if (type != 0) {
		cerr << "Dynamics::get_leaders_position() called by follower!" << endl;
		exit(EXIT_FAILURE);
	}

	State ret_val;
	int index = traveled_distance.size()-1;

	if (s > 0.0) {
		while (index>0) {
			if (std::abs(traveled_distance[index]-s) < 500*imr::constants::eps) break;
			index--;
		}
		if(index < state_history_fdif.size()){
		ret_val = state_history_fdif[index];
		}else{
		ret_val = state_history_fdif[0];
		ret_val.EA(2) = 5;
		}
	
	} else {
		
		ret_val = state_history_fdif[0];
		ret_val.X(0) += cos(ret_val.EA(2))*s;
		ret_val.X(1) += sin(ret_val.EA(2))*s;
	}
	
	return ret_val;
}

/**
 * Writes the simulation results to m-file (Matlab, Octave).
 */
void Dynamics::write_to_mfile()
{
	const int modulo = (int)(SAMPLE_T/DELTA_T);
	int counter;

	ofstream out(log_file);
	if (out.fail()) {
		std::cerr << "The file " << log_file << " cannot be opened." << std::endl;
		exit(EXIT_FAILURE);
	}
	
    out << "t = " << total_time << ";\n";
    out << "delta_t = " << DELTA_T << ";\n";
    out << "t_sim = 0:" << SAMPLE_T << ":" << total_time << ";\n";
	
	vector<State>::iterator st_it;
	vector<Control>::iterator ctr_it;
	
	counter = 0;
	out << "X = [ ";
	st_it = state_history.begin();
	while (st_it != state_history.end()) {
		if (counter % modulo == 0) out << st_it->X(0) << " ";
		counter ++;
		st_it ++;
	}
	out << "];\n";
	
	counter = 0;
	out << "Y = [ ";
	st_it = state_history.begin();
	while (st_it != state_history.end()) {
		if (counter % modulo == 0) out << st_it->X(1) << " ";
		counter ++;
		st_it ++;
	}
	out << "];\n";
	
	counter = 0;
	out << "Z = [ ";
	st_it = state_history.begin();
	while (st_it != state_history.end()) {
		if (counter % modulo == 0) out << st_it->X(2) << " ";
		counter ++;
		st_it ++;
	}
	out << "];\n";
	
	counter = 0;
	out << "Wx = [ ";
	st_it = state_history.begin();
	while (st_it != state_history.end()) {
		if (counter % modulo == 0) out << st_it->W(0) << " ";
		counter ++;
		st_it ++;
	}
	out << "];\n";
	
	counter = 0;
	out << "Wy = [ ";
	st_it = state_history.begin();
	while (st_it != state_history.end()) {
		if (counter % modulo == 0) out << st_it->W(1) << " ";
		counter ++;
		st_it ++;
	}
	out << "];\n";
	
	counter = 0;
	out << "Wz = [ ";
	st_it = state_history.begin();
	while (st_it != state_history.end()) {
		if (counter % modulo == 0) out << st_it->W(2) << " ";
		counter ++;
		st_it ++;
	}
	out << "];\n";
	
	counter = 0;
	out << "Rx = [ ";
	st_it = state_history.begin();
	while (st_it != state_history.end()) {
		if (counter % modulo == 0) out << st_it->EA(0) << " ";
		counter ++;
		st_it ++;
	}
	out << "];\n";
	
	counter = 0;
	out << "Ry = [ ";
	st_it = state_history.begin();
	while (st_it != state_history.end()) {
		if (counter % modulo == 0) out << st_it->EA(1) << " ";
		counter ++;
		st_it ++;
	}
	out << "];\n";
	
	counter = 0;
	out << "Rz = [ ";
	st_it = state_history.begin();
	while (st_it != state_history.end()) {
		if (counter % modulo == 0) out << st_it->EA(2) << " ";
		counter ++;
		st_it ++;
	}
	out << "];\n";
	out << "V = [ ";
	st_it = state_history.begin();
	while (st_it != state_history.end()) {
		if (counter % modulo == 0) out << sqrt(st_it->V(0)*st_it->V(0) + st_it->V(1)*st_it->V(1)) << " ";
		counter ++;
		st_it ++;
	}
	out << "];\n";

	counter = 1;
	out << "f = [ ";
	ctr_it = control_history.begin();
	while (ctr_it != control_history.end()) {
		if (counter % modulo == 0) out << ctr_it->f << " ";
		counter ++;
		ctr_it ++;
	}
	out << "];\n";
	
	out << "Mx = [ ";
	for(int i=0; i<control_history.size(); i+=modulo) {
		double max = 0.0;
		for(int j=0; j<modulo; j++) {
			if (abs(control_history[i+j].m(0)) > abs(max)) max = control_history[i+j].m(0);
		}
		out << max*DELTA_T << " ";
	}
	out << "];\n";

	out << "My = [ ";
	for(int i=0; i<control_history.size(); i+=modulo) {
		double max = 0.0;
		for(int j=0; j<modulo; j++) {
			if (abs(control_history[i+j].m(1)) > abs(max)) max = control_history[i+j].m(1);
		}
		out << max*DELTA_T << " ";
	}
	out << "];";

	out << "Mz = [ ";
	for(int i=0; i<control_history.size(); i+=modulo) {
		double max = 0.0;
		for(int j=0; j<modulo; j++) {
			if (abs(control_history[i+j].m(2)) > abs(max)) max = control_history[i+j].m(2);
		}
		out << max*DELTA_T << " ";
	}
	out << "];\n";

	out.close();

	cout << "File with results stored in: " << log_file << endl;
}
