#include "quadrotor.h"

#include <iostream>
#include <Eigen/Geometry>

#include "../settings.h"

using namespace std;
using namespace imr::dynamics;

/**
 * Constructor of the object
 * @param initial_X - initial position of the quadcopter
 * @param initial_V - initial velocity of the quadcopter
 * @param initial_OMEGA - initial angular velocity of the quadcopter
 * @param initial_R - initial rotation matrix of the quadcopter
 * */
Quadrotor::Quadrotor(const Vector & initial_X, const Vector & initial_V, const Vector & initial_OMEGA,
																const Matrix & initial_R, const int _type) : type(_type)
{	
	traveled_fix = 0;
	/// Quadrotor constants initialization
	J << 0.082,0.0,   0.0,
	     0.0,  0.0845,0.0,
	     0.0,  0.0,   0.1377;

	invJ << 12.1951, 0.0,    0.0,
	         0.0,   11.8343, 0.0,
	         0.0,    0.0,    7.2622;

	g << 0,0, imr::constants::g;
	m = 4.34;
    
	/// Initialization of the quadrotor
	state.X = Vector(initial_X);
	state.V = Vector(initial_V);
	state.W = Vector(initial_OMEGA);
	state.R = Matrix(initial_R);
	state.EA = get_euler_angles(state.R); //state.R.eulerAngles(2,1,0);
    
	state_history.push_back(state);
	state_history_fdif.push_back(state);
	
	/// Total flight time
	total_time = 0.0;
}

/**
 * Simulates quadrotor dynamics.
 * @param time - discrete time interval (number of sample)
 * @param f - total thrust
 * @param mx - moment around x body-fixed asis
 * @param my - moment around y body-fixed asis
 * @param mz - moment around z body-fixed asis
 * */
void Quadrotor::move(const int time, const double f, const Vector moments)
{
		// Debuging output
		#ifdef DEBUG_QUADROTOR_DYNAMICS
		std::cout << "Input:" << std::endl;
		std::cout << "Time: " << time << std::endl;
		std::cout << "Delta t: " << DELTA_T << std::endl;
		std::cout << "f: " << f << std::endl;
		std::cout << "Moments: (" << moments.transpose() << ")'" << std::endl;
		#endif

	state = State();

	/// Actual position vector
	/// Acceleration		
	state.V = state_history[time-1].R*Eigen::Vector3d::UnitZ();
	state.V *= f/m;
	state.V.noalias() -= g;
	state.V *= DELTA_T;
	state.V.noalias() += state_history[time-1].V;

		// Debuging output
		#ifdef DEBUG_QUADROTOR_DYNAMICS
		std::cout << "Calculated velocity: (" << state.V.transpose() << ")'" << std::endl;
		#endif
	
	state.X = state.V*DELTA_T;
	state.X.noalias() += state_history[time-1].X;
		// Debuging output
		#ifdef DEBUG_QUADROTOR_DYNAMICS
		std::cout << "Calculated position: (" << state.X.transpose() << ")'" << std::endl;
		#endif

	/// Actual angular velocity vector
	Vector crossProd = moments;
	crossProd.noalias() -= state_history[time-1].W.cross(J*state_history[time-1].W);
	
	state.W = invJ*crossProd;
	state.W *= DELTA_T;
	state.W.noalias() += state_history[time-1].W;
		// Debuging output
		#ifdef DEBUG_QUADROTOR_DYNAMICS
		std::cout << "Calculated omega: (" << state.W.transpose() << ")'" << std::endl;
		#endif
		
	/// Actual rotation matrix
	state.R = state_history[time-1].R * calculate_Rw(state.W);
		// Debuging output
		#ifdef DEBUG_QUADROTOR_DYNAMICS
		std::cout << "Calculated R:\n" << state.R << std::endl;
		#endif
	
	/// Actual euler angles
		state.EA = get_euler_angles(state.R); //state.R.eulerAngles(2,1,0);
	
	/// Add actual state to the history vector
    state_history.push_back(state);
    state_history_fdif.push_back(state);
    
    /// Add actual control inputs to the history vector
    #ifdef LOG_CONTROL
        control_history.push_back(Control(f, moments));
    #endif

    /// Tracking travelled distance for the leader
    if (type == 0) {
    	const double s0 = traveled_distance.back();
		traveled_distance.push_back((sqrt(sqr(state.V(0))+sqr(state.V(1)))*DELTA_T)+s0);
		traveled_distance_fdif.push_back((sqrt(sqr(state.V(0))+sqr(state.V(1)))*DELTA_T)+s0-traveled_fix);
    }
}

/**
 * Returns new rotation matrix (correction).
 * @param omega - angular velocity vector
 * */
const Matrix Quadrotor::calculate_Rw(const Vector & omega)
{
	Matrix ret_val;
	if (omega.norm() < constants::eps) {
		ret_val = Eigen::Matrix3d::Identity();
	} else {
		Vector axis(omega);
		axis.normalize();
		
		double theta = omega.norm()*DELTA_T;
		
		/// Rotation matrix around general axis w
		ret_val = Eigen::AngleAxis<double>(theta, axis);
	}
	return ret_val;
}


/**
 * Simulates quadrotor dynamics with constant inputs.
 * @param t - duration of the time interval with constant inputs
 * @param f - total thrust
 * @param mx - moment around x body-fixed asis
 * @param my - moment around y body-fixed asis
 * @param mz - moment around z body-fixed asis
 * */
void Quadrotor::static_control_move(const double t, const double f, const Vector moments)
{
	int discr_time = state_history.size();
	int max_discr_time = discr_time + (int)(t/DELTA_T); // there is an initial state so +1 is not needed

	while (discr_time < max_discr_time) {
		move(discr_time, f, moments);
		discr_time ++;
	}
	total_time += t;

		// Debuging output
		#ifdef DEBUG_QUADROTOR_MOVE
		std::cout << "static_control_move " << std::endl;
		std::cout << "Control: " << std::endl;
		std::cout << f << std::endl;
		std::cout << "(" << moments.transpose() << ")'" << std::endl;
		std::cout << "Time: " << t << std::endl;
		std::cout << "Final position: (" << state_history.back().X.transpose() << ")'" << std::endl;
		std::cout << "Final velocity: (" << state_history.back().V.transpose() << ")'" << std::endl;
		std::cout << "Final angular velocity: (" << state_history.back().W.transpose() << ")'" << std::endl;
		std::cout << "Final Euler angles: (" << state_history.back().EA.transpose() << ")'" << std::endl;
		#endif
}
/**
 * Executes the rotation maneuver (leader).
 * @param ag - angle of the rotation maneuver.
 * @param formationCenter - center of the formation.
 * @param diamter - rotation maneuver circle diameter.
 * */
void Quadrotor::setRot(const double ag, const Vector formationCenter, double diameter){
	Matrix Rot;
	double angle = state.EA(2) + ag;
	Rot << cos(ag),-sin(ag),0,
	     sin(ag),cos(ag),0,
	     0,0,1;
	Vector V0;
	V0 << 0,0,0;
	state.V = V0;
	state.R = state.R*Matrix(Rot);
	state.EA = get_euler_angles(state.R);
	state.X(0) = formationCenter(0) + diameter*cos(angle);
	state.X(1) = formationCenter(1) + diameter*sin(angle);
	state_history_fdif.clear();
	state_history_fdif.push_back(state);
	state_history.push_back(state);
}
/**
 * Executes the rotation maneuver (follower).
 * @param ag - angle of the rotation maneuver of the leader.
 * */
void Quadrotor::setRotFol(const double ag){
	Matrix Rot;
	double angle = ag;
	Rot << cos(angle),-sin(angle),0,
	     sin(angle),cos(angle),0,
	     0,0,1;	
	state.R = Matrix(Rot);
	state.EA = get_euler_angles(state.R);
	state_history.push_back(state);
}
/**
 * Function returning the Euler angles computed from a rotation matrix
 * @param R - a rotation matrix
 * */
Vector Quadrotor::get_euler_angles(const Matrix & R)
{
	double phi1 = 0; // Roll  (around x-axis)
	double phi2 = 0; // Pitch (around y-axis)
	double phi3 = 0; // Yaw   (around z-axis)

	//   0 1 2
	// 0 a b c
	// 1 d e f
	// 2 g h i

	if (R(2,2) < -1.0-constants::eps || R(2,2) > 1.0+constants::eps) {
		cerr << "Wrong matrix coeffitient!" << endl;
		exit(1);
	}

	if (abs(R(2,0) - 1) < constants::eps) {
		// R(3,1) ~ 1
		phi1 = atan2(-R(0,1), -R(0,2));
		phi2 = constants::pi/2.0;
		phi3 = 0;
	} else if (abs(R(2,0) + 1) < constants::eps) {
		// R(3,1) ~ -1
		phi1 = atan2(R(0,1), R(0,2));
		phi2 = -constants::pi/2.0;
		phi3 = 0;
	} else {
		// R(3,1) \in (-1,1)
		phi1 = atan2(R(2,1)/cos(phi2), R(2,2)/cos(phi2));
		phi2 = -asin(R(2,0));
		phi3 = atan2(R(1,0)/cos(phi2), R(0,0)/cos(phi2));
	}

	Vector ret_val(phi1, phi2, phi3);
	return ret_val;
}
