#include "robot_follower.h"

#include "../utils.h"
#include "../world.h"
#include "../solver/solver.h"

#include <cmath>
#include <stdio.h>
#include <fstream>
#include <iostream>

using namespace std;
using namespace imr::robot;
using namespace imr::defects;

/**
 * Constructor of the class Robot.
 * @param id            - id of the robot 0 - leader, 1..n - followers
 * @param fs			- formation specification parameters
 * @param n             - number of constant duration control inputs
 * @param d             - follower defect (see utils.h for more details)
 * @param initial_X     - initial position of the quadrotor
 * @param initial_V     - initial velocity of the quadrotor
 * @param initial_OMEGA - initial angular velocity of the quadrotor
 * @param initial_R     - initial rotation matrix of the quadrotor
 * @param filename      - file to store the simulation results
 * */
RobotFollower::RobotFollower(const int _id, const FSpec & _fs, const int _n, const double dt, FollowerDefect d,
		const Vector & initial_X, const Vector & initial_V, const Vector & initial_OMEGA, const Matrix & initial_R, const char * filename) :
			Robot::Robot(_id, _n, initial_X, initial_V, initial_OMEGA, initial_R, filename), fs(_fs), fsoriginal(_fs), defect_time(dt), defect(d), fault_vt(0.15),
			fault_vn(-0.05), fault_ag(0.0)
{
	traveled_s = 0.0;
	time = 0.0;
	ra = 0.60;
	rs = 0.0;
}

/**
 * Run the movement of the formation
 */
void RobotFollower::run(const Trajectory leadertraj_)
{	
	const Vector curr = robot_dynamics.get_state().X;	
	const Vector angles = robot_dynamics.get_state().EA;	
	const Vector vels = robot_dynamics.get_state().V;
	leadertraj = leadertraj_;
	if(leadertraj[0].vt == 0){
		
	cout << "	Follower " << id << ":" << endl;
	cout << "		Position: (" << curr.transpose() << ")',"<<", ag = "<< angles[2] <<", vx = "<< vels[0] <<", vy = "<< vels[1] <<", vz = "<< vels[2] <<", t = " << time <<" p = "<< fs.p<< " q = "<< fs.q<<endl;
	ol.add_str(time,1);  
	robot_move(leadertraj[0]);
			
		time += TIME_EXECUTE;
			return;
		
	}
	Solver * solver;
	const OptVec ub(limits.agMax, limits.vtMax, limits.vnMax, limits.cMax, TIME_EXECUTE);
	const OptVec lb(limits.agMin, limits.vtMin, limits.vnMin, limits.cMin, TIME_EXECUTE);


	cout << "	Follower " << id << ":" << endl;

	cout << "		Position: (" << curr.transpose() << ")',"<<endl;

	switch (defect) {
	// Normal function
	case FolDefNone:
			init();
				cout << "		Trajectory size = " << traj.size() << ", n = " << n << endl;
				cout << "		Starting solver ..." << endl;
			tm.start();
			solver = new Solver(curr, traj.size(), id, n, traj, lb, ub, ra, rs, fs);
			tm.end();
			if (solver->get_status() == 2) {
				cerr << "		Obtained solver status [" << solver->get_status() << "] ..." << endl;
				cout << "	fitness: " << solver->get_fitness() << endl;
				cout << "			...following preceding trajectory." << endl;
			} else {
				cout << "		Obtained solver status [" << solver->get_status() << "] ..." << endl;
				cout << "	fitness: " << solver->get_fitness() << endl;
				cout << "			...using the result." << endl;
				traj = solver->get_solution();		


		
			}
				cout << "		vt = " << traj[0].vt << ", vn = " << traj[0].vn << ", c = " << traj[0].c << ", t = " << traj[0].t << "\n" << endl;



			ol.add_str(time,solver->get_fitness());
			robot_move(traj[0]);
			traj.erase(traj.begin());

			delete solver;
		break;

	// The follower stays on the initial position
	case FolDefStatic:
		cout << "		Defect STATIC !" << endl;
		robot_move(OptVec(0.0, 0.0,0.0,0.0,TIME_EXECUTE));
		break;

	// The follower stops after time 'defect_time'
	case FolDefTimedStop:
		if (time < defect_time) {
			init();
				cout << "		Trajectory size = " << traj.size() << ", n = " << n << endl;
				cout << "		Starting solver ..." << endl;
			tm.start();
			solver = new Solver(curr, traj.size(), id, n, traj, lb, ub, ra, rs, fs);
			tm.end();
			if (solver->get_status() == 2) {
				cerr << "		Obtained solver status [" << solver->get_status() << "] ..." << endl;
				cout << "			...following preceding trajectory." << endl;
			} else {
				cout << "		Obtained solver status [" << solver->get_status() << "] ..." << endl;
				cout << "			...using the result." << endl;
				traj = solver->get_solution();
			}
				cout << "		vt = " << traj[0].vt << ", vn = " << traj[0].vn << ", c = " << traj[0].c << ", t = " << traj[0].t << "\n" << endl;
			robot_move(traj[0]);
			traj.erase(traj.begin());

			delete solver;
		} else {
			cout << "		Defect TIMED STOP !" << endl;
			robot_move(OptVec(fault_ag, fault_vt,fault_vn,0.0,TIME_EXECUTE));
		}
		break;

	// The follower keeps by the last control input after time 'defect_time'
	case FolDefTimedFault:
		if (time < defect_time) {
			init();
				cout << "		Trajectory size = " << traj.size() << ", n = " << n << endl;
				cout << "		Starting solver ..." << endl;
			tm.start();
			solver = new Solver(curr, traj.size(), id, n, traj, lb, ub, ra, rs, fs);
			tm.end();
			if (solver->get_status() == 2) {
				cerr << "		Obtained solver status [" << solver->get_status() << "] ..." << endl;
				cout << "			...following preceding trajectory." << endl;
			} else {
				cout << "		Obtained solver status [" << solver->get_status() << "] ..." << endl;
				cout << "			...using the result." << endl;
				traj = solver->get_solution();
			}
				cout << "		vt = " << traj[0].vt << ", vn = " << traj[0].vn << ", c = " << traj[0].c << ", t = " << traj[0].t << "\n" << endl;

			robot_move(traj[0]);
			traj.erase(traj.begin());

			delete solver;
		} else {
			cout << "		Defect TIMED STOP !" << endl;
			robot_move(traj[0]);
		}
		break;
	}

	time += TIME_EXECUTE;
}

/**
 * Initialize the robot
 */
void RobotFollower::init()
{
	const Trajectory t = world.leader->init_ctr(n,fs.p,fs.q);
	int bound = min(n,world.leader->traj_size());
	for(int i=traj.size();i<bound;i++) {
		traj.push_back(t[i]);
		
	}
}

/**
 * Moves the robot by given control inputs.
 * @param ctr - control inputs
 * */
void RobotFollower::robot_move(const OptVec & ctr)
{
	const Vector V0 = robot_dynamics.get_state().V;
	robot_dynamics.move_ctr(ctr);

	/// Average velocity of the follower
	Vector V = robot_dynamics.get_state().V;
	V.noalias() += V0;
	V /= 2.0;

	/// Transformation to the leader's track
	const double cl = ctr.c/(fs.q*ctr.c + 1);
	const double vtl = sqrt(sqr(V(0))+sqr(V(1)))/(1-fs.q*cl);

	traveled_s += vtl*ctr.t;
}
/**
 * Returns information about leader's rotation maneuvers.
 * @param traj     - leader's  trajectory
 * @param nsamples - number of samples

 * */
std::vector<double> RobotFollower::get_rots(OptVec * traj, const int nsamples){
std::vector<double> rots;
for (int i = 0; i < n; i++) {
	for(int j = 0; j < nsamples; j++) {
 		if (abs(leadertraj[i].ag)>imr::constants::eps){
			rots.push_back(leadertraj[i].ag);		
		}else{
			rots.push_back(0);		
		}	
	}
}
	return rots;
}

/**
 * Returns sampled trajectory.
 * @param traj     - pointer to the control inputs
 * @param len      - length of array ctr
 * @param nsamples - count of samples (first n ctr inputs)
 * @param msamples - count of samples (last m ctr inputs)
 * @param es       - export switch: 0=no export; 1=text only; 2=matlab only; 3=text and matlab
 * @param filename - name of the file to store the results to
 * @return Samples - samples
 * */
Samples RobotFollower::sample_trajectory(OptVec * traj, const int len, const int nsamples, const int msamples, const int es, const char * filename)
{
	Samples ret_val;
    const State curr = robot_dynamics.get_state();



    double yaw;
    if (abs(leadertraj[0].ag)  > imr::constants::eps){
	yaw = curr.EA(2) + leadertraj[0].ag;
    }else{
   	yaw = curr.EA(2);
    }
    Point pos;
    vector2point(curr.X, pos, time);

    /// Time of one part of the segment
    double sample_t;
    /// YAW angle increment
    double yaw_inc;

    //ret_val.push_back(Point(pos));

    /// Loop over all given control inputs


	for (int i = 0; i < n; i++) {
	
        /// Apply control
        if (traj[i].t > constants::eps) {
            sample_t = traj[i].t/nsamples;
            if (abs(traj[i].c) > constants::eps) {
		if (leadertraj[i].ag != 0.0){
			yaw += leadertraj[i].ag;	
		}
                yaw_inc = traj[i].c*traj[i].vt*sample_t;
                for(int j = 0; j < nsamples; j++) {
		
		if(es == 1){
}
                    pos(0) += (sin(yaw+yaw_inc) - sin(yaw))/traj[i].c;
                    pos(1) -= (cos(yaw+yaw_inc) - cos(yaw))/traj[i].c;
                    pos(2) += traj[i].vn*sample_t;
                    pos(3) += sample_t;

                    yaw += yaw_inc;
		

        			const double cl = traj[i].c/(fs.q*traj[i].c + 1);
        			const double vtl = traj[i].vt/(1-fs.q*cl);
		
                    ret_val.push_back(Point(pos));
		
                }
            } else {
		if (leadertraj[i].ag != 0.0){
			yaw += leadertraj[i].ag;	
		}
                for(int j = 0; j < nsamples; j++) {
			if(es == 1){
}
                    pos(0) += traj[i].vt*cos(yaw)*sample_t;
                    pos(1) += traj[i].vt*sin(yaw)*sample_t;
                    pos(2) += traj[i].vn*sample_t;
                    pos(3) += sample_t;
			

                    ret_val.push_back(Point(pos));
                }
            }
        }
    }

    if (es == 1 || es == 3) {
        cout << "------ ROBOT EDGE SAMPLES ------" << endl;
        for(int i = 0; i < ret_val.size(); i++) {
            cout << "Sample #" << i+1 << ": (" << ret_val[i].transpose() << ")" << endl;
		i+=3;
        }
        cout << "------ ROBOT EDGE SAMPLES ------" << endl;
    }

    if (es == 2 || es == 3) {
        ofstream out(filename);
        out << "sX = [ ";
        for(int i = 0; i < ret_val.size(); i++) {
            out << (ret_val[i])(0) << " ";
        }
        out << "];\n";

        out << "sY = [ ";
        for(int i = 0; i < ret_val.size(); i++) {
            out << (ret_val[i])(1) << " ";
        }
        out << "];\n";

        out << "sZ = [ ";
        for(int i = 0; i < ret_val.size(); i++) {
            out << (ret_val[i])(2) << " ";
        }
        out << "];\n";

        out << std::flush;
        out.close();

        cout << "Matlab file '" << filename << "' created!" << endl;
    }

    return ret_val;
}

/**
 * Returns expected state at given time 't'.
 */
const State RobotFollower::state_estimation_t(const double t)
{
	State ret_val(robot_dynamics.get_state());
	double t_to_go = t - time;

	double yaw_inc = 0.0;
	double tt = 0.0;
	int i = 0;

	switch (defect) {
	// Normal function
	case FolDefNone:
		if (!traj.empty()) {
			while (t_to_go > imr::constants::eps) {
				if (i >= traj.size()) break;

				tt = min(t_to_go, traj[i].t);

				if (abs(traj[i].c) > constants::eps) {
					yaw_inc = traj[i].c*traj[i].vt*tt;
					ret_val.X(1) -= (cos(ret_val.EA(2)+yaw_inc) - cos(ret_val.EA(2)))/traj[i].c;
					ret_val.X(0) += (sin(ret_val.EA(2)+yaw_inc) - sin(ret_val.EA(2)))/traj[i].c;
					ret_val.X(2) += traj[i].vn*tt;
					ret_val.EA(2) += yaw_inc;
				} else {
					ret_val.X(0) += traj[i].vt*cos(ret_val.EA(2))*tt;
					ret_val.X(1) += traj[i].vt*sin(ret_val.EA(2))*tt;
					ret_val.X(2) += traj[i].vn*tt;
				}

				t_to_go -= tt;
				i++;
			}
		} else {
			ret_val = robot_dynamics.get_state();
		}
		break;

	// The follower stays on the initial position
	case FolDefStatic:
		ret_val = robot_dynamics.get_state();
		break;

	// The follower stops after time 'defect_time'
	case FolDefTimedStop:
		if (time < defect_time && !traj.empty()) {
			while (t_to_go > imr::constants::eps) {
				tt = min(t_to_go, traj[i].t);

		        if (abs(traj[i].c) > constants::eps) {
		        	yaw_inc = traj[i].c*traj[i].vt*tt;
					ret_val.X(0) += (sin(ret_val.EA(2)+yaw_inc) - sin(ret_val.EA(2)))/traj[i].c;
					ret_val.X(1) -= (cos(ret_val.EA(2)+yaw_inc) - cos(ret_val.EA(2)))/traj[i].c;
					ret_val.X(2) += traj[i].vn*tt;
					ret_val.EA(2) += yaw_inc;
		        } else {
					ret_val.X(0) += traj[i].vt*cos(ret_val.EA(2))*tt;
					ret_val.X(1) += traj[i].vt*sin(ret_val.EA(2))*tt;
					ret_val.X(2) += traj[i].vn*tt;
		        }

				t_to_go -= tt;
				i++;
			}
		} else {
			ret_val = robot_dynamics.get_state();
			/*ret_val.X(0) += fault_vt*cos(ret_val.EA(2))*t_to_go;
			ret_val.X(1) += fault_vt*sin(ret_val.EA(2))*t_to_go;
			ret_val.X(2) += fault_vn*t_to_go;*/
		}
		break;

	// The follower keeps by the last control input after time 'defect_time'
	case FolDefTimedFault:
		if (time < defect_time && !traj.empty()) {
			while (t_to_go > imr::constants::eps) {
				tt = min(t_to_go, traj[i].t);

		        if (abs(traj[i].c) > constants::eps) {
		        	yaw_inc = traj[i].c*traj[i].vt*tt;
					ret_val.X(0) += (sin(ret_val.EA(2)+yaw_inc) - sin(ret_val.EA(2)))/traj[i].c;
					ret_val.X(1) -= (cos(ret_val.EA(2)+yaw_inc) - cos(ret_val.EA(2)))/traj[i].c;
					ret_val.X(2) += traj[i].vn*tt;
					ret_val.EA(2) += yaw_inc;
		        } else {
					ret_val.X(0) += traj[i].vt*cos(ret_val.EA(2))*tt;
					ret_val.X(1) += traj[i].vt*sin(ret_val.EA(2))*tt;
					ret_val.X(2) += traj[i].vn*tt;
		        }

				t_to_go -= tt;
				i++;
			}
		} else {
			if (!traj.empty()) {
				if (abs(traj[0].c) > constants::eps) {
					yaw_inc = traj[0].c*traj[0].vt*tt;
					ret_val.X(0) += (sin(ret_val.EA(2)+yaw_inc) - sin(ret_val.EA(2)))/traj[0].c;
					ret_val.X(1) -= (cos(ret_val.EA(2)+yaw_inc) - cos(ret_val.EA(2)))/traj[0].c;
					ret_val.X(2) += traj[0].vn*tt;
					ret_val.EA(2) += yaw_inc;
				} else {
					ret_val.X(0) += traj[0].vt*cos(ret_val.EA(2))*tt;
					ret_val.X(1) += traj[0].vt*sin(ret_val.EA(2))*tt;
					ret_val.X(2) += traj[0].vn*tt;
				}
			} else {
				ret_val = robot_dynamics.get_state();
			}
		}
		break;
	}

	return ret_val;
}

/**
 * Returns expected position at given time 't'.
 */
const Vector RobotFollower::get_position_t(const double t)
{
	Vector ret_val;

	if (std::abs(t-time) < imr::constants::eps) {
		ret_val = robot_dynamics.get_state().X;
	} else {
		ret_val = state_estimation_t(t).X;
	}

	return ret_val;
}
/**
 * Exexcutes the rotation maneuver
 * @param ag   - new angle of the follower
 * @param newp - new p formation parameter
 * @param newq - new q foramtion parameter
 * */
const void RobotFollower::rotate(double ag, double newp, double newq){
traj.clear();
fs.p = newp;
fs.q = newq;

robot_dynamics.setRotFol(ag);


}

const Vector RobotFollower::get_position()
{
Vector ret_val;
State state;
state = robot_dynamics.get_state();
	ret_val[0]=state.X(0);
	ret_val[1]=state.X(1);
	return ret_val;
}
/**
 * Tasks performed after leaders main loop is finished
 */
void RobotFollower::post_run()
{
	char str[3];
	sprintf (str, "%d", id);
	string objLog("../scripts/fol");
	objLog += str;
	objLog += "Obj.m";

	write_results();
	ol.write(objLog.data());
}
