#include "robot_leader.h"

#include "../world.h"
#include "../settings.h"
#include "../solver/solver.h"

#include <iostream>
#include <fstream>
#include <cmath>

using namespace std;
using namespace imr::robot;
using namespace imr::settings;

/**
 * Constructor of the class Robot.
 * @param id            - id of the robot 0 - leader, 1..n - followers
 * @param n             - number of constant duration control inputs
 * @param initial_X     - initial position of the quadcopter
 * @param initial_V     - initial velocity of the quadcopter
 * @param initial_OMEGA - initial angular velocity of the quadcopter
 * @param initial_R     - initial rotation matrix of the quadcopter
 * @param filename      - file to store the simulation results
 * */
RobotLeader::RobotLeader(const int _id, const int _n, const Vector & initial_X, const Vector & initial_V, const Vector & initial_OMEGA,
    	   const Matrix & initial_R, const char * filename) : Robot(_id, _n, initial_X, initial_V, initial_OMEGA, initial_R, filename)
{
	traveled_s = 0.0;
	rot_s = 0;
	time = 0.0;
	rotations = 0.0;
	diameter = 0;
}

/**
 * Run the movement of the formation
 */
void RobotLeader::run()
{
	init();

	const OptVec ub(limits.agMax, 0.8*limits.vtMax, limits.vnMax, 0.8*limits.cMax, LEADER_TIME_STEP_MAX);
	const OptVec lb(limits.agMin, limits.vtMin, limits.vnMin, 0.8*limits.cMin, LEADER_TIME_STEP_MIN);
	double dist;
	double maxdist = 0;
	int fol1;
	int fol2;
	double tmp;
	double newp;
	double newq;
	FSpec fs;
	Vector f1;
	Vector f2;
	int lc = 0;
	double d2g = 0.0;	
	Vector folpos;
	Vector formationcenter;	
		for (int i=0; i<world.followers.size(); i++) {
				f1 = world.followers[i]->get_position();
				for(int j=0; j<world.followers.size(); j++){
					f2 = world.followers[j]->get_position();
					tmp = distance(f1(0),f1(1),f2(0),f2(1));
					
					if (tmp > maxdist){
						maxdist = tmp;
						fol1 = i;
						fol2 = j;					
					}			
				}
				
			}
			f1 = world.followers[fol1]->get_position();			
			f2 = world.followers[fol2]->get_position();
			formationcenter(0) =  (f1(0)+f2(0))/2;
			formationcenter(1) =  (f1(1)+f2(1))/2;
			diameter = (maxdist/2)+0.3;
			
	sample_trajectory(&traj[0], traj.size(), 10, 50, 3, "../scripts/initialLoader.m");




	Solver * solver;

	while (!traj.empty()) {
		const Vector curr = robot_dynamics.get_state().X;
		d2g = world.map->goal_distance(curr);

			cout << " ==== Loop " << ++lc << " (" << d2g << " m remaining) ==== " << endl;
			cout << "	Position: (" << curr.transpose() << ")', t = " << time << endl;
			cout << "	Trajectory size = " << traj.size() << ", n = " << n << endl;

			cout << "Starting solver ..." << endl;

		tm.start();
		solver = new Solver(curr, traj.size(), id, n, traj, lb, ub, ra, rs, FSpec());
		tm.end();

		if (solver->get_status() == 2) {
			cerr << "Obtained solver status [" << solver->get_status() << "] ..." << endl;
			cout << "	fitness: " << solver->get_fitness() << endl;
			cout << "	...following preceding trajectory." << endl;
		} else {
			cout << "Obtained solver status [" << solver->get_status() << "] ..." << endl;
			cout << "	fitness: " << solver->get_fitness() << endl;
			cout << "	...using the result." << endl;
			traj = solver->get_solution();
		}
			cout << "vt = " << traj[0].vt << ", vn = " << traj[0].vn << ", c = " << traj[0].c << ", t = " << traj[0].t <<  ", ag = " << traj[0].ag <<"\n\n\n" << endl;

				cout << "\n	------ Running followers:" << endl;
				for (int i=0; i<world.followers.size(); i++) {
					world.followers[i]->run(traj);
				}
				cout << "	------ followers finished ......\n\n" << endl;

		ol.add_str(time, solver->get_fitness());

		State st = robot_dynamics.get_state();
	if (abs(traj[0].ag)  > imr::constants::eps){
			if(time != 0){
			Trajectory tmptraj;
			tmptraj.push_back(OptVec(0.0, 0.0,0.0,0.0,0.2));
			for(int i = 0; i<3; i++){ 
				cout<< "Braking for rotation" << endl;
				for (int j=0; j<world.followers.size(); j++) {
						world.followers[j]->run(tmptraj);
					}
			
				robot_move(tmptraj[0]);			
				time += TIME_EXECUTE;
			}}	
			vector<Vector> foldesired (world.followers.size());
			formationcenter(0) = 0;
			formationcenter(1) = 0;
			State state;
			state = robot_dynamics.get_state();
			double angle = state.EA(2) + traj[0].ag;
				
			for (int i=0; i<world.followers.size(); i++) {
				f1 = world.followers[i]->get_position();
				for(int j=0; j<world.followers.size(); j++){
					f2 = world.followers[j]->get_position();
					tmp = distance(f1(0),f1(1),f2(0),f2(1));
					if (tmp > maxdist){
						maxdist = tmp;
						fol1 = i;
						fol2 = j;					
					}			
				}
				
			}
			f1 = world.followers[fol1]->get_position();			
			f2 = world.followers[fol2]->get_position();
			formationcenter(0) =  (f1(0)+f2(0))/2;
			formationcenter(1) =  (f1(1)+f2(1))/2;
			diameter = (maxdist/2)+0.3;

			for (int i=0; i<world.followers.size(); i++){
				folpos = world.followers[i]->get_position();

				
				fs = world.followers[i]->fspec();
					cout<< fs.p<< " "<< fs.q<<endl;
			foldesired[i](0) = state.X(0)-fs.p*cos(state.EA(2)) - fs.q*sin(state.EA(2));
			foldesired[i](1) = state.X(1)+fs.q*cos(state.EA(2)) - fs.p*sin(state.EA(2));	
			
			cout<< foldesired[i](0)<< " "<< foldesired[i](1)<<" "<< i<<endl;	
			}	
	
			robot_dynamics.setRot(traj[0].ag, formationcenter, diameter);
		
			state = robot_dynamics.get_state();
			for (int i=0; i<world.followers.size(); i++){
			
			newp = (state.X(0)-foldesired[i](0))*cos(state.EA(2)) + (state.X(1)-foldesired[i](1))*sin(state.EA(2));
			newq = (foldesired[i](1)-state.X(1))*cos(state.EA(2)) + (state.X(0)-foldesired[i](0))*sin(state.EA(2));	
			
		
			world.followers[i]->rotate(angle, newp, newq);	
				
			}

			rotations += traj[0].ag;
			traveled_s_fdif = traveled_s;
			rot_s = time;		
				
		}
		robot_move(traj[0]);
		time += TIME_EXECUTE;
		reorganize();

		delete solver;
	}

	cout << " ==== Loop terminated (" << d2g << " m remaining) ==== " << endl;

	robot_dynamics.write_to_mfile();
	for (int i=0; i<world.followers.size(); i++) {
		world.followers[i]->post_run();
	}
	ol.write("../scripts/leaderObj.m");
	tm.write_results("scripts/leader_time_measurement.m");
	
}

/**
 * Reorganize the trajectory (after the first control input is used)
 */
void RobotLeader::reorganize()
{
	traj.erase(traj.begin());
	
	
	if (traj.size() > n-1) {
	repeat:
		if (traj[n-1].t > TIME_EXECUTE/2.0) {
			traj.insert(traj.begin()+n-1, OptVec(traj[n-1].ag,traj[n-1].vt,traj[n-1].vn,traj[n-1].c,TIME_EXECUTE));
			traj[n].t -= TIME_EXECUTE;
			if (traj[n].t < TIME_EXECUTE/2.0) traj.erase(traj.begin()+n);
		} else {
			traj.erase(traj.begin()+n-1);
			if (traj.size() > n-1) goto repeat;
		}
	}
}

/**
 * Initialize the robot
 */
double RobotLeader::distance(double x1, double y1, double x2, double y2){
double ret_val = sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2));
return ret_val;
}
void RobotLeader::init()
{
	adapt_limitations();

	Trajectory tmp(world.map->get_initial_trajectory());
		if (tmp.empty()) {
			cerr << "RobotLeader::init() - Obtained epmty vector!" << endl;
			exit(EXIT_FAILURE);
		}

	// First n - constant duration ctr inputs
	int index = 0;
	for (int i = 0; i < n; i++) {
		traj.push_back(OptVec(tmp[index].ag, tmp[index].vt, tmp[index].vn, tmp[index].c, TIME_EXECUTE));
		tmp[index].t -= TIME_EXECUTE;
		while (tmp[index].t < TIME_EXECUTE/2.0 && !tmp.empty()) index ++;
	}
	// Additional m - varying duration ctr inputs
	for (int i = index; i < tmp.size(); i++) {
		traj.push_back(OptVec(tmp[i]));
	}
}

/**
 * Adapt leaders limitation to the rest of the formation
 */
void RobotLeader::adapt_limitations()
{
	if (world.followers.empty()) {
		cerr << "RobotLeader::adapt_limitations() -> 'No followers registered!'" << endl;
	}

	ra = 0.5;
	rs = 0.6;

	for(int i=0; i<world.followers.size(); i++) {
		/// Ensure formation safe flight
		const double ra_tmp = sqrt(sqr(world.followers[i]->q())+sqr(world.followers[i]->r()))-FOLLOWER_CONSTR_EUC;
		const double rs_tmp = 1.75*world.followers[i]->get_rs()+sqrt(sqr(world.followers[i]->q())+sqr(world.followers[i]->r()));
		if (ra < ra_tmp) ra = ra_tmp; // max
		if (rs < rs_tmp) rs = rs_tmp; // max

		/// Ensure trajectory feasibility for all formation members
		const double cMax_tmp = world.followers[i]->cMax()/(1+world.followers[i]->q()*world.followers[i]->cMax());
		if (limits.cMax > cMax_tmp) limits.cMax = cMax_tmp; // min
		const double cMin_tmp = world.followers[i]->cMin()/(1-world.followers[i]->q()*world.followers[i]->cMin());
		if (limits.cMin < cMin_tmp) limits.cMin = cMin_tmp; // max
	}

	for(int i=0; i<world.followers.size(); i++) {
		const double vMin_tmp = world.followers[i]->vtMin()/(1+world.followers[i]->q()*limits.cMax);
		if (limits.vtMin < vMin_tmp) limits.vtMin = vMin_tmp; // max
		const double vMax_tmp = world.followers[i]->vtMax()/(1+world.followers[i]->q()*limits.cMax);
		if (limits.vtMax > vMax_tmp) limits.vtMax = vMax_tmp; // min
	}

	// ra = 0.50;
	// rs = 0.60;

		#ifdef LIST_LEADERS_LIMITATIONS
		cout << "Leaders limitations are:" << endl;
		cout << "	Maximal c  = " << limits.cMax << endl;
		cout << "	Minimal c  = " << limits.cMin << endl;
		cout << "	Maximal vt = " << limits.vtMax << endl;
		cout << "	Minimal vt = " << limits.vtMin << endl;
		cout << "	        ra = " << ra << endl;
		cout << "	        rs = " << rs << endl;
		#endif
}

/**
 * Moves the robot by given control inputs.
 * @param ctr - control inputs
 * */
void RobotLeader::robot_move(const OptVec & ctr)
{
	const Vector V0 = robot_dynamics.get_state().V;
	traveled_s = robot_dynamics.move_ctr(ctr);

	/// Average velocity
	Vector V = robot_dynamics.get_state().V;
	V.noalias() += V0;
	V /= 2.0;

	ctr_history.push_back(ctr);
	ctr_dist.push_back(sqrt(sqr(V(0))+sqr(V(1)))*ctr.t);
}
/**
 * Checks the leader's trajectory
 * @param traj - leader's trajectory
 * @param len  - trajectory length
 * */
int RobotLeader::check_trajectory(OptVec * traj, const int len){
	int ret_val = 0;
	for (int i = 0; i < len; i++){
		if(abs(traj[i].ag)  > imr::constants::eps && traj[i].t != 0.2)	{
			ret_val = 1;
		}
	}
	return ret_val;
}
/**
 * Returns sampled trajectory.
 * @param traj           - pointer to the control inputs
 * @param len            - length of array ctr
 * @param nsamples       - count of samples (first n ctr inputs)
 * @param msamples       - count of samples (last m ctr inputs)
 * @param es             - export switch: 0=no export; 1=text only; 2=matlab only; 3=text and matlab
 * @param filename       - name of the file to store the results to
 * @return vector<Point> - sample_cnt states along the edge
 * */
Samples RobotLeader::sample_trajectory(OptVec * traj, const int len, const int nsamples, const int msamples, const int es, const char * filename)
{
	Samples ret_val;

    double yaw;
    const State curr = robot_dynamics.get_state();
    Point pos;
    vector2point(curr.X, pos, time);

   
    if(abs(traj[0].ag)  > imr::constants::eps){
	yaw = curr.EA(2);
	pos(0) -= diameter*cos(yaw);
	pos(1) -= diameter*sin(yaw);
	yaw += traj[0].ag;
	pos(0) += diameter*cos(yaw);
	pos(1) += diameter*sin(yaw);
	pos(3) += 0.6;
    }else{
    yaw = curr.EA(2);
    }

    /// Time of one part of the segment
    double sample_t;
    /// YAW angle increment
    double yaw_inc;

    /// Loop over all given control inputs
    for (int i = 0; i < n; i++) {	
        /// Apply control
        if (traj[i].t > constants::eps) {
            sample_t = traj[i].t/nsamples;
            if (abs(traj[i].c) > constants::eps) {
                yaw_inc = traj[i].c*traj[i].vt*sample_t;	


 			
                for(int j = 0; j < nsamples; j++) {
		    pos(0) += (sin(yaw+yaw_inc) - sin(yaw))/traj[i].c;
                    pos(1) -= (cos(yaw+yaw_inc) - cos(yaw))/traj[i].c;
                    pos(2) += traj[i].vn*sample_t;
                    pos(3) += sample_t;
                    yaw += yaw_inc;
		
if (abs(traj[i].ag)  > imr::constants::eps && j == 0 && i!= 0){
				pos(0) -= diameter*cos(yaw);
				pos(1) -= diameter*sin(yaw);
				yaw += traj[i].ag;
				pos(0) += diameter*cos(yaw);
				pos(1) += diameter*sin(yaw);				
				pos(3) += 0.6;						
			}

                    ret_val.push_back(pos);
                }
            } else {
		
 			
                for(int j = 0; j < nsamples; j++) {
                    pos(0) += traj[i].vt*cos(yaw)*sample_t;
                    pos(1) += traj[i].vt*sin(yaw)*sample_t;
                    pos(2) += traj[i].vn*sample_t;
                    pos(3) += sample_t;
	
if (abs(traj[i].ag)  > imr::constants::eps && j == 0 && i!= 0){
				pos(0) -= diameter*cos(yaw);
				pos(1) -= diameter*sin(yaw);
				yaw += traj[i].ag;
				pos(0) += diameter*cos(yaw);
				pos(1) += diameter*sin(yaw);				
				pos(3) += 0.6;					
			}

                    ret_val.push_back(pos);
                }
            }
        }
    }

    for (int i = n; i < len; i++) {
        /// Apply control
        if (traj[i].t > constants::eps) {
            sample_t = traj[i].t/msamples;
            if (abs(traj[i].c) > constants::eps) {
                yaw_inc = traj[i].c*traj[i].vt*sample_t;

                for(int j = 0; j < msamples; j++) {
	
                    pos(0) += (sin(yaw+yaw_inc) - sin(yaw))/traj[i].c;
                    pos(1) -= (cos(yaw+yaw_inc) - cos(yaw))/traj[i].c;
                    pos(2) += traj[i].vn*sample_t;
                    pos(3) += sample_t;
                    yaw += yaw_inc;
		
if (abs(traj[i].ag)  > imr::constants::eps && j == 0 && i!= 0){
				
				pos(0) -= diameter*cos(yaw);
				pos(1) -= diameter*sin(yaw);
				yaw += traj[i].ag;
				pos(0) += diameter*cos(yaw);
				pos(1) += diameter*sin(yaw);					
				pos(3) += 0.6;					
			}

                    ret_val.push_back(pos);
                }
            } else {	
			
 			
                for(int j = 0; j < msamples; j++) {
		
                    pos(0) += traj[i].vt*cos(yaw)*sample_t;
                    pos(1) += traj[i].vt*sin(yaw)*sample_t;
                    pos(2) += traj[i].vn*sample_t;
                    pos(3) += sample_t;
if (abs(traj[i].ag)  > imr::constants::eps && j == 0 && i!= 0){
				pos(0) -= diameter*cos(yaw);
				pos(1) -= diameter*sin(yaw);
				yaw += traj[i].ag;
				pos(0) += diameter*cos(yaw);
				pos(1) += diameter*sin(yaw);				
				pos(3) += 0.6;
										
			}

                    ret_val.push_back(pos);
                }
            }
        }
    }

    if (es == 1 || es == 3) {
        cout << "------ ROBOT EDGE SAMPLES ------" << endl;
        for(int i = 0; i < ret_val.size(); i++) {
            cout << "Sample #" << i+1 << ": (" << ret_val[i].transpose() << ") "  << endl;
			i+=20;
		        }
        cout << "------ ROBOT EDGE SAMPLES ------" << endl;
    }

    if (es == 2 || es == 3) {
        ofstream out(filename);
        out << "sX = [ ";
        for(int i = 0; i < ret_val.size(); i++) {
            out << (ret_val[i])(0) << " ";
        }
        out << "];\n";

        out << "sY = [ ";
        for(int i = 0; i < ret_val.size(); i++) {
            out << (ret_val[i])(1) << " ";
        }
        out << "];\n";

        out << "sZ = [ ";
        for(int i = 0; i < ret_val.size(); i++) {
            out << (ret_val[i])(2) << " ";
        }
        out << "];\n";

        out << std::flush;
        out.close();

        cout << "Matlab file '" << filename << "' created!" << endl;
    }

    return ret_val;
}

/**
 * Returns difference between desired follower's position
 * and its actual position.
 */


const FDiff RobotLeader::get_formation_difference(Samples smp, FSpec fs, FSpec f_orig,std::vector<double> rots, int vypis)
{
	FDiff ret_val;
	int j = 0;
	double j_watch = 0.0;

	double dp = 0.0;
	double dq = 0.0;
	double dr = 0.0;
	double delta_t = 0.02;
	double sfi = (traveled_s-traveled_s_fdif) - fs.p;
	int stop = 0;
	State desired;

	for (int i=0;i<smp.size();i++) {
		if(stop == 0){
		sfi += delta_t*traj[j].vt;	
		j_watch += delta_t;
		if (abs(j_watch - traj[j].t) < imr::constants::eps) {
			j_watch = 0.0;
			j++;
			if (j == traj.size()) {
				// cerr << "RobotLeader::get_formation_difference(): Trajectory out of bounds!" << endl;
				break;
			}
		}
		if (i < smp.size()-1) delta_t = (smp[i+1])(3) - (smp[i])(3);

		if (std::abs(sfi - traveled_s) < 500*imr::constants::eps) {
			/// Use current leader's state
			desired = robot_dynamics.get_state();
		} else if (sfi < traveled_s) {
			/// Use leader's past state
			desired = robot_dynamics.get_leaders_state(sfi, vypis);
			
		} else {
			/// Perform a future state estimation
			desired = state_estimation_s(sfi);
		}

		desired.X(0) -= fs.q*sin(desired.EA(2));
		desired.X(1) += fs.q*cos(desired.EA(2));
		desired.X(2) += fs.r;
		if(abs(rots[i]) > imr::constants::eps){
			stop = 1;
		}
		if (desired.EA(2) == 5){
			return ret_val;			
		}
		
		}

		if ((abs(desired.EA(2))-rotations) < imr::constants::eps) {
			
			// No rotation needed
			dp = abs(desired.X(0)-(smp[i])(0));
			dq = abs(desired.X(1)-(smp[i])(1));
			dr = abs(desired.X(2)-(smp[i])(2));
		} else {
			// Rotate back |EA(2)| rad around z-axis in order to easily determine
			// the difference in normal (q) and tangential (p) direction
				Matrix Rz;
			Rz << cos(-desired.EA(2)+rotations), -sin(-desired.EA(2)+rotations), 0.0, sin(-desired.EA(2)+rotations), cos(-desired.EA(2)+rotations), 0.0, 0.0, 0.0, 1.0;
			desired.X = Rz*desired.X;
			Vector sample((smp[i])(0), (smp[i])(1), (smp[i])(2));
			sample = Rz*sample;

if (vypis == 3){cout<<"x = "<<desired.X(0)<< " y = "  << desired.X(1) << " samplex "<<  sample(0) <<"  sampley = " << sample(1) <<endl;}
			dp = abs(desired.X(0)-sample(0));
			dq = abs(desired.X(1)-sample(1));
			dr = abs(desired.X(2)-sample(2));
		}

		if (dp > P_TRESHOLD && dp > ret_val.pMax) ret_val.pMax = dp;
		if (dp > P_TRESHOLD) ret_val.pSum += dp;

		if (dq > Q_TRESHOLD && dq > ret_val.qMax) ret_val.qMax = dq;
		if (dq > Q_TRESHOLD) ret_val.qSum += dq;

		if (dr > R_TRESHOLD && dr > ret_val.rMax) ret_val.rMax = dr;
		if (dr > R_TRESHOLD) ret_val.rSum += dr;
		if(stop == 1){
			return ret_val;		
		}		
		if(stop > 0){
			stop++;		
		}

	}
		
	return ret_val;
}
Trajectory RobotLeader::init_ctr(const int n, const double p, const double q)
{
	Trajectory ret_val;

	int j = 0;
	double ts = traveled_s - p;
	double tjt = traj[j].t;

	for(int i=0;i<n;i++) {
		ret_val.push_back(traj[j]);

		if (ts < -constants::eps) {
			ret_val.back().vn = 0.0;
			ret_val.back().c = 0.0;
			ret_val.back().t = TIME_EXECUTE;
		} else {
			if (ts + constants::eps < traveled_s) {
				/// Control history
				int index = ctr_dist.size()-1;
				//while (index > -1 && ts < ctr_dist[index]) {
				//	index --;cout<< index<<"  "<<ts <<endl;
				//}cout<< " 2 "<<ts <<endl;
				
				ret_val.back().vn = ctr_history[index].vn;
				ret_val.back().c = ctr_history[index].c;
				ret_val.back().t = TIME_EXECUTE;
			} else {
				/// Control inputs
				double s = traveled_s;
				int index = 0;
				while (true) {
					if (s+traj[index].vt*traj[index].t < ts) {
						s += traj[index].vt*traj[index].t;
						index ++;
					} else {
						break;
					}
				}
				ret_val.back().vn = traj[index].vn;
				ret_val.back().c = traj[index].c;
				ret_val.back().t = TIME_EXECUTE;
			}

			if (abs(ret_val.back().c) > imr::constants::eps) {
				const double vt_L = ret_val.back().vt;
				const double c_L = ret_val.back().c;
				ret_val.back().c = c_L/(1-q*c_L);
				ret_val.back().vt = (c_L/ret_val.back().c)*vt_L;
			}
		}
		tjt -= TIME_EXECUTE;
		ts += ret_val.back().vt*ret_val.back().t;
		if (tjt < imr::constants::eps) {

			j++;
			tjt = traj[j].t;
		}
	}

	return ret_val;
}

const int RobotLeader::traj_size() const
{
	int ret_val = 0;

	for(int i=0;i<(int)traj.size();i++) {
		ret_val += (int)(traj[i].t/TIME_EXECUTE);
	}
		// cout << "Traj size: " << ret_val << endl;
	return ret_val;
}
const Trajectory RobotLeader::get_traj()
{
	
	Trajectory ret_val;
	ret_val = traj;
	return ret_val;
}
