 /**
  * CFSQP solver
  * @author Zdenek Kasl (kaslzden@fel.cvut.cz)
  * 
  * Based on CTU FEE solver wrapper by Jan Svab (7.4.2010).
  */
#include "solver.h"
#include <cmath>

#include "../world.h"
#include "../settings.h"

using namespace imr::settings;
using namespace std;
using namespace imr::time;
const int CP_STRUCT_DC = (int)(sizeof(struct imr::utils::quadrotor_opt)/sizeof(double)); // Control Point Struct Doubles Count

extern "C" {
	int x_is_new;
	double objrep,objeps,gLgeps;
	#include "../cfsqp/include/cfsqpusr.h"
}

/**
 * Computes value of the objective functions.
 * @param nparam - number of values in the optimalised vector; cd->ctr_len
 * @param j      - index of objective function
 * @param x      - pointer to the optimalised vector
 * @param fj     - pointer to the return value
 * @param cd     - pointer to void, here used to pass the Solver class pointer
 */
void obj(int nparam,int j, double *x, double *fj, void *cd)
{		Uint32 start_ticks;
			Uint32 end_ticks;
 //   if (j != 1) {
     //   cerr << "Something is very wrong in calling function obj() -> parameter 'j' ! Got " << j << endl;
  //      *fj = 0.0;
  //      return;
  //  }

    Solver * so = (Solver *)cd;
    std::vector<double> rots;

    if (so->id == 0) {
    	/// Leader
    	const Samples pts = world.leader->sample_trajectory((OptVec *)x, so->ctr_len, LEADER_N_SAMPLES, LEADER_M_SAMPLES, 0, "");

    	if (pts.empty()) {
    		cerr << "Solver::obj() -> PTS EMPTY!" << endl;
		}

    	/// Trajectory evaluation parameters:
    	// Time Of Flight
    	double tof = 0.0;
    	// Velocity difference
    	double vdiff = 0.0;
    	// Curvature difference
    	double cdiff = 0.0;
    	// Curvature penalty
    	double cpen = 0.0;
    	// Obstacle penalty
    	double obst = 0.0;
    	// Distance to goal penalty
    	double d2g = 0.0;
	// Rotation maneuver penalty
	double turn = 0.0;
		// Travelled distance
		double tdist = 0.0;

		double avg_vt = 0.0;
		double avg_vn = 0.0;
		double avg_c = 0.0;

		for(int i = 0; i < so->ctr_len; i++) {
			tof += ((OptVec *)x)[i].t;
			cpen += abs(((OptVec *)x)[i].c);
			avg_vt += ((OptVec *)x)[i].vt;
			avg_vn += ((OptVec *)x)[i].vn;
			avg_c += ((OptVec *)x)[i].c;
			tdist += sqrt(sqr(((OptVec *)x)[i].vt)+sqr(((OptVec *)x)[i].vn))*((OptVec *)x)[i].t;
			if( abs(((OptVec *)x)[i].ag) > imr::constants::eps){
				if(abs(((OptVec *)x)[i].ag) < 0.7){
					turn += log(10*abs(((OptVec *)x)[i].ag) +1);				
				}else{
					turn += (0.1*abs(((OptVec *)x)[i].ag) + 2.00944);				
				}	

				tof += 0.6;		
			}
		}

		avg_vt /= so->ctr_len;
		avg_vn /= so->ctr_len;
		avg_c /= so->ctr_len;

		for(int i = 0; i < so->ctr_len; i++) {
			vdiff += abs(((OptVec *)x)[i].vt-avg_vt)+abs(((OptVec *)x)[i].vn-avg_vn);
			cdiff += abs(((OptVec *)x)[i].c-avg_c);
		}

		d2g = world.map->goal_distance(Vector(pts.back()(0),pts.back()(1),pts.back()(2)));

		double curr = numeric_limits<double>::max();
		double tmp = numeric_limits<double>::max();
		for (int i = 0; i < pts.size(); i++) {
			tmp = world.map->obstacle_distance(pts[i], world.leader->get_time());
			if (tmp < curr) curr = tmp;
		}
		obst = sqr(min(0.0, (curr - so->rs)/(curr - so->ra)));

		*fj = LEADER_OBJ_TOF*tof +
			  LEADER_OBJ_VDI*vdiff +
			  LEADER_OBJ_CDI*cdiff +
			  LEADER_OBJ_CPE*cpen +
			  LEADER_OBJ_OBS*obst +
			  LEADER_OBJ_D2G*d2g +
			  LEADER_OBJ_TURN*turn +
			  LEADER_OBJ_TDI*tdist;


    } else {
    	/// Followers
    	const Samples pts = world.followers[so->id-1]->sample_trajectory((OptVec *)x, so->ctr_len, FOLLOWER_N_SAMPLES, 0.0, 0, "");
	rots = world.followers[so->id-1]->get_rots((OptVec *)x, FOLLOWER_N_SAMPLES);


    	if (pts.empty()) {
    		cerr << "Solver::obj() -> PTS EMPTY!" << endl;
		}

    	/// Trajectory evaluation parameters:
    	// Formation difference		
    	const FDiff fdif = world.leader->get_formation_difference(pts, so->fs, world.followers[so->id-1]->get_fsorig(), rots, 0);
    	// Obstacle avodidance
    	double obst = 0.0;
    	// Formation member collision avoidance
    	double foll = 0.0;
    	// Velocity difference
    	double vdiff = 0.0;
    	// Curvature difference
    	double cdiff = 0.0;

		double avg_vt = 0.0;
		double avg_vn = 0.0;
		double avg_c = 0.0;

		double curr1 = 10^50; // obst
		double curr2 = 10^50; // foll
		double tmp = 10^50;
		for (int i = 0; i < pts.size(); i++) {
			tmp = world.map->obstacle_distance(pts[i],0.0);
			if (tmp < curr1) curr1 = tmp;

			for (int j=0;j<world.followers.size(); j++) {
				if (j+1 != so->id) {
					tmp = dist2D(pts[i], world.followers[j]->get_position_t((pts[i])(3))) - 0.5; // TODO: ROBOT SIZE !!!
					if (tmp < curr2) curr2 = tmp;
				}
			}
		}
		if (!pts.empty()) {
			if (curr1 < so->ra) curr1 = so->ra+imr::constants::eps;
			obst = sqr(min(0.0, (curr1 - so->rs)/(curr1 - so->ra)));

			if (curr2 < 0.1/*FA*/) curr2 = /*TODO FA*/0.1+imr::constants::eps;
			foll = sqr(min(0.0, (curr2 - 1.0)/(curr2 - 0.1))); //TODO: FS, FA -> formation collision specification
		} else {
			obst = 0.0;
			foll = 0.0;
		}
		for (int i = 0; i < so->ctr_len; i++) {
			avg_vt += ((OptVec *)x)[i].vt;
			avg_vn += ((OptVec *)x)[i].vn;
			avg_c += ((OptVec *)x)[i].c;
		}

		avg_vt /= so->ctr_len;
		avg_vn /= so->ctr_len;
		avg_c /= so->ctr_len;

		for(int i = 0; i < so->ctr_len; i++) {
			vdiff += abs(((OptVec *)x)[i].vt-avg_vt)+abs(((OptVec *)x)[i].vn-avg_vn);
			cdiff += abs(((OptVec *)x)[i].c-avg_c);
		}

		*fj = FOLLOWER_OBJ_PS*(fdif.pSum) +
			  FOLLOWER_OBJ_PM*fdif.pMax +
			  FOLLOWER_OBJ_QS*(fdif.qSum) +
			  FOLLOWER_OBJ_QM*fdif.qMax +

			  FOLLOWER_OBJ_RS*(fdif.rSum) +
			  FOLLOWER_OBJ_RM*fdif.rMax +
			  FOLLOWER_OBJ_VDI*vdiff +

			  FOLLOWER_OBJ_CDI*cdiff +
			  FOLLOWER_OBJ_OBS*obst +
			  FOLLOWER_OBJ_FOL*foll;

		}

}

/**
 * Computes value of the constraints.
 * @param nparam - number of values in the optimalised vector; cd->ctr_len
 * @param j      - index of constraint function
 * @param x      - pointer to the optimalised vector
 * @param gj     - pointer to the return value
 * @param cd     - pointer to void, here used to pass the Solver class pointer
 */
void constr(int nparam, int j, double *x, double *gj, void *cd){

    Solver * so = (Solver *)cd;

    Samples pts;
    double curr;
    double curr2;
    double tmp;
    std::vector<double> rots;

    if (so->id == 0) {
    	pts = world.leader->sample_trajectory((OptVec *)x, so->ctr_len, LEADER_N_SAMPLES, LEADER_M_SAMPLES, 0, "");
    } else {
    	pts = world.followers[so->id-1]->sample_trajectory((OptVec *)x, so->ctr_len, FOLLOWER_N_SAMPLES, 0.0, 0, "");	
	rots = world.followers[so->id-1]->get_rots((OptVec *)x, FOLLOWER_N_SAMPLES);
    }
    switch (j) {
        case 1:
        	if (pts.empty()) {
        		cerr << "Solver::constr() -> PTS EMPTY!" << endl;
        		*gj = 10.0;
        	} else {
        			if (so->id == 0) {
					curr = numeric_limits<double>::max();
					tmp = numeric_limits<double>::max();
					for (int i = 0; i < pts.size(); i++) {
						tmp = world.map->obstacle_distance(pts[i], world.leader->get_time());
						if (tmp < curr) curr = tmp;
					}
					*gj = so->ra - curr;
								for (int i = 0; i < LEADER_M_SAMPLES; i++) {
			if (abs(((OptVec *)x)[i].ag) > imr::constants::eps && ((OptVec *)x)[i].t != 0.2) {
				int check = world.leader->check_trajectory((OptVec *)x, so->ctr_len);
				if(check == 1){
				*gj = 10.0;}
			}}
					if(j > 2){
					cout<< "gj1= "<< *gj<< endl;}
        		} else {
    				if (pts.empty()) {
    					*gj = 10.0;
    	        		cerr << "Solver::constr() -> PTS EMPTY!" << endl;
    				} else {
						curr = numeric_limits<double>::max();
						curr2 = numeric_limits<double>::max();
						tmp = numeric_limits<double>::max();
						for (int i = 0; i < pts.size(); i++) {
							tmp = world.map->obstacle_distance(pts[i], world.leader->get_time());
							if (tmp < curr) curr = tmp;

							for (int j=0;j<world.followers.size(); j++) {
								if (j+1 != so->id) {
									tmp = dist2D(pts[i], world.followers[j]->get_position_t((pts[i])(3))) - 0.5; // TODO: ROBOT SIZE !!!
									if (tmp < curr2) curr2 = tmp;
								}
							}
						}
						*gj = min(so->ra - curr, 0.1 - curr2/*TODO RA*/);
						//*gj = max(so->ra - curr-0.35, -10.0/*TODO RA*/);
			if(j > 2){
					cout<< "gj1= "<< *gj<<"prvni" <<so->ra - curr-0.35<<  "druha "<<  0.10 - curr2<< "  "<< curr2 <<endl;}
	
    				}
        		}
        	}
        	break;

        case 2:
        	if (so->id == 0) {
				if (pts.empty()) {
					*gj = 10.0;
	        		cerr << "Solver::constr() -> PTS EMPTY!" << endl;
				} else {
					*gj = world.map->goal_distance(Vector(pts.back()(0),pts.back()(1),pts.back()(2))) - DISTANCE_TO_GOAL;
		
				}
        	} else {
						
					
				if (pts.empty()) {
					*gj = 10.0;
	        		cerr << "Solver::constr() -> PTS EMPTY!" << endl;
				} else {
					const FDiff fdif = world.leader->get_formation_difference(pts, so->fs, world.followers[so->id-1]->get_fsorig(), rots, 0);
	        		*gj = max(fdif.qMax - FOLLOWER_CONSTR_EUC, fdif.pMax - FOLLOWER_CONSTR_EUC);	

				}
	
        	}
            break;

        default:
            cerr << "Something is very wrong in calling function constr() -> parameter 'j' !" << endl;
            *gj = -0.0;
            break;
    }
}

/**
 * Generates boundary values for the solver.
 */
void generate_bounds(double * bu, double * bl, const OptVec & _ub, const OptVec & _lb, const int nbounds, const int len)
{
	//cout << "nbounds " << nbounds << endl;
	//cout << "len " << len << endl;
	const int n = min(nbounds,len);

    for(int i = 0; i < n; i++) {
        for(int j = 0; j < (int)CP_STRUCT_DC; j++) {
            if (j == (int)(CP_STRUCT_DC - 1)) {
                bl[i*CP_STRUCT_DC+j] = TIME_EXECUTE;
                bu[i*CP_STRUCT_DC+j] = TIME_EXECUTE;
            } else {
                bl[i*CP_STRUCT_DC+j] = ((double*)&_lb)[j];
                bu[i*CP_STRUCT_DC+j] = ((double*)&_ub)[j];
            }
        }
    }

	for(int i = n; i < len; i++) {
		for(int j = 0; j < CP_STRUCT_DC; j++) {
			bl[i*CP_STRUCT_DC+j] = ((double*)&_lb)[j];
			bu[i*CP_STRUCT_DC+j] = ((double*)&_ub)[j];
		}
	}
}

/**
 * Function calling the cfsqp solver library.
 * @param so - pointer to the Solver.
 */
const int call_solver(Solver *so)
{
	const int nparam = so->ctr_len*CP_STRUCT_DC;
	const int nf     = 1; // number of objective functions (param j in obj())

	const int nfsr   = 0;
    
    // if changing nineq* it is necessary to change switch in function constr! (see the manual)
    // nonlinear constraints
	const int nineqn = 2;
	const int nineq  = 2;

    // linear constraints
	const int neqn   	= 0;
    const int neq    	= 0;

	const int ncsrl  	= 0;
	const int ncsrn  	= 0;
	int mesh_pts[1]		= {0};
	const int mode		= 200;

    const int iprint	= 0;

	const int miter	= 500;

	// Used status instead (solver.h) int inform; // output information
	const double bigbnd       	= 1.0e100;  //numeric_limits<double>::max();
	const double solver_eps   	= 1.0e-12; // 4.44089209850063e-16
	const double epseqn       	= 1.0e-6;  // 1e-6
	const double udelta       	= 0.0e0;

	double *bl = (double *)calloc(nparam,sizeof(double));
	double *bu = (double *)calloc(nparam,sizeof(double));
	generate_bounds(bu, bl, so->ub, so->lb, so->n, so->ctr_len);

    double *f = (double *)calloc((nf+1), sizeof(double));
    double *lambda = (double *)calloc(nineq+neq+nf+nparam,sizeof(double));
    double *g = (double *)calloc(nineq+neq+1,sizeof(double));
    
    cfsqp(nparam,nf,nfsr,nineqn,nineq,neqn,neq,ncsrl,ncsrn,mesh_pts,
    	  mode,iprint,miter,&so->status,bigbnd,solver_eps,epseqn,udelta,bl,bu,(double *)(&(so->trajectory[0])),
          f,g,lambda,obj,constr,grobfd,grcnfd,(void *)so);

    // cout << " [" << inform << "]" << endl;

    free((void *)bl);
    free((void *)bu);
    free((void *)f);
    free((void *)g);
    free((void *)lambda);

    return 0;
}

/**
 * Returns fitness of the trajectory (objective function result)
 */
double Solver::get_fitness()
{

	double ret_val;
obj(ctr_len, 1, (double *)&trajectory[0],&ret_val, (void *)this);
	
	return ret_val;
}

/**
 * Constructor of class Solver.
 */
Solver::Solver(const Vector & _sp, const int _ctr_len, const int _id, const int _n, const Trajectory & traj,
			const OptVec _lb, const OptVec _ub, const double _ra, const double _rs, const FSpec & _fs) : start_point(_sp), ctr_len(_ctr_len), id(_id), n(_n), trajectory(traj),
            lb(_lb), ub(_ub), ra(_ra), rs(_rs), fs(_fs), status(-1)
{
 	 call_solver(this);
}
