#ifndef _QUADROTOR_UTILS_H_
#define _QUADROTOR_UTILS_H_

#include <cmath>
#include <iostream>
#include <Eigen/Core>

#include "settings.h"
using namespace imr::settings;

namespace imr {
    namespace constants {
        const double eps = 1e-6;
        const double  pi = 3.14159265358979323846;
        const double   g = 9.81;
    }
    
    namespace defects{
		enum FollowerDefect {FolDefNone, FolDefStatic, FolDefTimedStop, FolDefTimedFault};
		// FolDefNone       = normal function
		// FolDefStatic     = the follower stays on the initial postion
		// FolDefTimedStop  = the follower stops after time t
		// FolDefTimedFault = the follower keeps by the last control input after time t
    }

    namespace utils {
        typedef Eigen::Matrix3d Matrix;
        typedef Eigen::Vector3d Vector;
        typedef Eigen::Vector4d Point;

        #define sqr(x) (double)((x)*(x))
        #define dist(a,b) sqrt(sqr(a(0)-b(0))+sqr(a(1)-b(1))+sqr(a(2)-b(2)))
		#define dist2D(a,b) sqrt(sqr(a(0)-b(0))+sqr(a(1)-b(1)))

        #define point2vector(p,v) v(0)=p(0);v(1)=p(1);v(2)=p(2)
        #define vector2point(v,p,t) p(0)=v(0);p(1)=v(1);p(2)=v(2);p(3)=t

		struct quadrotor_state {
			Vector X;
			Vector V;
			Vector W;
			Matrix R;
			Vector EA;

			quadrotor_state() : X(0,0,0), V(0,0,0), W(0,0,0), R(Eigen::Matrix3d::Identity()), EA(0,0,0) {};
			quadrotor_state(const quadrotor_state & st) : X(st.X), V(st.V), W(st.W), R(st.R), EA(st.EA) {};
			quadrotor_state(const Vector _X, const Vector _V, const Vector _W, const Matrix _R, const Vector _EA) :
																					X(_X), V(_V), W(_W), R(_R), EA(_EA) {};
		};
        
		struct quadrotor_control {
			double f;
			Vector m;
			
			quadrotor_control() : f(0), m(0,0,0) {};
			quadrotor_control(const quadrotor_control & qc) : f(qc.f), m(qc.m) {};
			quadrotor_control(const double _f, const Vector _m) : f(_f), m(_m) {};
			quadrotor_control(const double _f, const double _mx, const double _my, const double _mz) :
																					f(_f), m(_mx, _my, _mz) {};
		};

		struct quadrotor_limits {
			double agMax;
			double vtMax;
			double vnMax;
			double atMax;
			double anMax;
			double cMax;

			double agMin;
			double vtMin;
			double vnMin;
			double atMin;
			double anMin;
			double cMin;

			quadrotor_limits() : agMax(AG_MAX), vtMax(VT_MAX), vnMax(VN_MAX), atMax(AT_MAX), anMax(AN_MAX), cMax(CURV_MAX),
								 agMin(AG_MIN), vtMin(VT_MIN), vnMin(VN_MIN), atMin(AT_MIN), anMin(AN_MIN), cMin(CURV_MIN) {};
		};
		
		struct quadrotor_opt {	
			double ag;
			double vt;
			double vn;
			double c;
			double t;

			quadrotor_opt() : ag(0.0), vt(0.0), vn(0.0), c(0.0), t(0.0) {};
			quadrotor_opt(const quadrotor_opt & qo) : ag(qo.ag), vt(qo.vt), vn(qo.vn), c(qo.c), t(qo.t) {};
			quadrotor_opt(const double _ag, const double _vt, const double _vn, const double _c, const double _t) :
																					ag(_ag), vt(_vt), vn(_vn), c(_c), t(_t) {};
		};

		struct formation_spec {
			double p;
			double q;
			double r;

			formation_spec() : p(0.0), q(0.0), r(0.0) { };
			formation_spec(const formation_spec & fs) : p(fs.p), q(fs.q), r(fs.r) { };
			formation_spec(const double _p, const double _q, const double _r) : p(_p), q(_q), r(_r) { };
		};

		struct obstacle {	
		double nodes[8][3];
		double time;
		obstacle(double a[][3], double t){	
				for (int i = 0; i < 8; i++){
					for (int j = 0; j < 3; j++){	
						nodes[i][j] = a[i][j];
					}				
				}
				time = t;
			}
		void print() const{
		for (int i = 0; i < 8; i++){
					for (int j = 0; j < 3; j++){	
						std::cout<<nodes[i][j]<<" ";
						if(j == 2){
							std::cout<<""<<std::endl;						
						}
					}				
				}
		}	
		};
		struct formation_diff {
			double pMax;
			double pSum;

			double qMax;
			double qSum;

			double rMax;
			double rSum;

			formation_diff() : pSum(0.0), qSum(0.0), rSum(0.0), pMax(0.0), qMax(0.0), rMax(0.0) {};

			formation_diff(const formation_diff & d) : pSum(d.pSum), qSum(d.qSum), rSum(d.rSum),
					pMax(d.pMax), qMax(d.qMax), rMax(d.rMax) {};
		};
    }
}
#endif
